import os, datetime, sys
import numpy as np
import torch
import SimpleITK as sitk
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def checkPatchDivisibility(patch_size, number=16):
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (numpy.array): The patch size for checking.
        number (int, optional): The number to check divisibility for. Defaults to 16.

    Returns:
        bool: If all elements of array are divisible or not, after taking 2D patches into account.
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]
    # for 2D, don't check divisibility of first dimension
    elif patch_size_to_check[0] == 1:
        patch_size_to_check = patch_size_to_check[1:]
    if np.count_nonzero(np.remainder(patch_size_to_check, number)) > 0:
        return False

    # adding check to address https://github.com/mlcommons/GaNDLF/issues/53
    # there is quite possibly a better way to do this
    unique = np.unique(patch_size_to_check)
    if (unique.shape[0] == 1) and (unique[0] < number):
        return False
    return True


def get_date_time():
    """
    Get a well-parsed date string

    Returns:
        str: The date in format YYYY/MM/DD::HH:MM:SS
    """
    now = datetime.datetime.now().strftime("%Y/%m/%d::%H:%M:%S")
    return now


def get_unique_timestamp():
    """
    Get a well-parsed timestamp string to be used for unique filenames

    Returns:
        str: The date in format YYYYMMDD_HHMMSS
    """
    now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    return now


def get_filename_extension_sanitized(filename):
    """
    This function returns the extension of the filename with leading and trailing characters removed.
    Args:
        filename (str): The filename to be processed.
    Returns:
        str: The filename with extension removed.
    """
    _, ext = os.path.splitext(filename)
    # if .gz or .nii file is detected, always return .nii.gz
    if (ext == ".gz") or (ext == ".nii"):
        ext = ".nii.gz"
    return ext


def parse_version(version_string):
    """
    Parses version string, discards last identifier (NR/alpha/beta) and returns an integer for comparison.

    Args:
        version_string (str): The string to be parsed.

    Returns:
        int: The version number.
    """
    version_string_split = version_string.split(".")
    if len(version_string_split) > 3:
        del version_string_split[-1]
    return int("".join(version_string_split))


def version_check(version_from_config, version_to_check):
    """
    This function checks if the version of the config file is compatible with the version of the code.

    Args:
        version_from_config (str): The version of the config file.
        version_to_check (str): The version of the code or model to check.

    Returns:
        bool: If the version of the config file is compatible with the version of the code.
    """
    version_to_check_int = parse_version(version_to_check)
    min_ver = parse_version(version_from_config["minimum"])
    max_ver = parse_version(version_from_config["maximum"])
    if (min_ver > version_to_check_int) or (max_ver < version_to_check_int):
        sys.exit(
            "Incompatible version of GaNDLF detected ("
            + str(version_to_check_int)
            + ")"
        )

    return True


def checkPatchDimensions(patch_size, numlay):
    """
    This function checks the divisibility of a numpy array or integer for architectural integrity

    Args:
        patch_size (numpy.array): The patch size for checking.
        number (int, optional): The number to check divisibility for. Defaults to 16.

    Returns:
        int: Largest multiple of 2 (less than or equal to numlay) that each element of patch size is divisible by to yield int >= 2
    """
    if isinstance(patch_size, int):
        patch_size_to_check = np.array(patch_size)
    else:
        patch_size_to_check = patch_size
    # for 2D, don't check divisibility of last dimension
    if patch_size_to_check[-1] == 1:
        patch_size_to_check = patch_size_to_check[:-1]

    if all(
        [x >= 2 ** (numlay + 1) and x % 2**numlay == 0 for x in patch_size_to_check]
    ):
        return numlay
    else:
        # base2 = np.floor(np.log2(patch_size_to_check))
        base2 = np.array([getBase2(x) for x in patch_size_to_check])
        remain = patch_size_to_check / 2**base2  # check that at least 1

        layers = np.where(remain == 1, base2 - 1, base2)
        return int(np.min(layers))


def getBase2(num):
    # helper for checkPatchDimensions (returns the largest multiple of 2 that num is evenly divisible by)
    base = 0
    while num % 2 == 0:
        num = num / 2
        base = base + 1
    return base


def get_array_from_image_or_tensor(input_tensor_or_image):
    """
    This function returns the numpy array from a tensor or image.
    Args:
        input_tensor_or_image (torch.Tensor or sitk.Image): The input tensor or image.
    Returns:
        numpy.array: The numpy array from the tensor or image.
    """
    if isinstance(input_tensor_or_image, torch.Tensor):
        return input_tensor_or_image.detach().cpu().numpy()
    elif isinstance(input_tensor_or_image, sitk.Image):
        return sitk.GetArrayFromImage(input_tensor_or_image)
    elif isinstance(input_tensor_or_image, np.ndarray):
        return input_tensor_or_image
    else:
        raise ValueError("Input must be a torch.Tensor or sitk.Image or np.ndarray")



import torch.nn as nn
import sys
import torchvision


def brainage(parameters):
    if parameters["model"]["dimension"] != 2:
        sys.exit("Brain Age predictions only works on 2D data")
    model = torchvision.models.vgg16(pretrained=True)
    model.final_convolution_layer = None
    # Freeze training for all layers
    for param in model.features.parameters():
        param.require_grad = False
    # Newly created modules have require_grad=True by default
    num_features = model.classifier[6].in_features
    features = list(model.classifier.children())[:-1]  # Remove last layer
    # features.extend([nn.AvgPool2d(1024), nn.Linear(num_features,1024),nn.ReLU(True), nn.Dropout2d(0.8), nn.Linear(1024,1)]) # RuntimeError: non-empty 2D or 3D (batch mode) tensor expected for input
    features.extend(
        [
            nn.Linear(num_features, 1024),
            nn.ReLU(True),
            nn.Dropout2d(0.8),
            nn.Linear(1024, 1),
        ]
    )
    model.classifier = nn.Sequential(*features)  # Replace the model classifier
    parameters["model"]["amp"] = False  # this is not yet implemented for vgg

    return model



import os
from GANDLF.compute import create_pytorch_objects
from GANDLF.parseConfig import parseConfig
from GANDLF.utils import version_check, load_model, optimize_and_save_model


def post_training_model_optimization(model_path, config_path):
    """
    CLI function to optimize a model for deployment.

    Args:
        model_path (str): Path to the model file.
        config_path (str): Path to the config file.

    Returns:
        bool: True if successful, False otherwise.
    """

    parameters = parseConfig(config_path, version_check_flag=False)
    (
        model,
        _,
        _,
        _,
        _,
        parameters,
    ) = create_pytorch_objects(parameters, device="cpu")

    main_dict = load_model(model_path, "cpu")
    version_check(parameters["version"], version_to_check=main_dict["version"])
    model.load_state_dict(main_dict["model_state_dict"])
    optimize_and_save_model(model, parameters, model_path, onnx_export=True)
    optimized_model_path = model_path.replace("pth.tar", "onnx")
    if not os.path.exists(optimized_model_path):
        print("Error while optimizing model.")
        return False
    return True



import torch.nn as nn
from GANDLF.models.seg_modules.Interpolate import Interpolate


class FCNUpsamplingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        scale_factor=2,
        interpolation_mode="trilinear",
    ):
        nn.Module.__init__(self)
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 1, "stride": 1, "padding": 0, "bias": True}

        self.interp_kwargs = {
            "size": None,
            "scale_factor": 2 ** (scale_factor - 1),
            "mode": interpolation_mode,
            "align_corners": True,
        }
        self.interpolate = Interpolate(interp_kwargs=self.interp_kwargs)

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)

    def forward(self, x):
        """[summary]

        [description]

        Extends:
        """
        x = self.interpolate(self.conv0(x))
        return x



import torch.nn as nn
import torch.nn.functional as F


class GlobalAveragePooling2D(nn.Module):
    def __init__(self):
        super(GlobalAveragePooling2D, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 4, x.size()
        B, C, W, H = x.size()

        # This is a temporary fix to make sure the size is an integer, not a tensor
        if isinstance(B, int):
            return F.avg_pool2d(x, (W, H)).view(B, C)
        else:
            return F.avg_pool2d(x, (W.item(), H.item())).view(B.item(), C.item())


class GlobalAveragePooling3D(nn.Module):
    def __init__(self):
        super(GlobalAveragePooling3D, self).__init__()

    def forward(self, x):
        assert len(x.size()) == 5, x.size()
        B, C, W, H, D = x.size()

        # This is a temporary fix to make sure the size is an integer, not a tensor
        if isinstance(B, int):
            return F.avg_pool3d(x, (W, H, D)).view(B, C)
        else:
            return F.avg_pool3d(x, (W.item(), H.item(), D.item())).view(
                B.item(), C.item()
            )



# -*- coding: utf-8 -*-
"""
Modified from https://github.com/pytorch/vision.git
"""

import torchvision
import torch.nn as nn
import torch.nn.functional as F

from .modelBase import ModelBase


def create_torchvision_model(modelname, pretrained=True, num_classes=2, dimensions=2):
    """
    Create a torchvision model with the given parameters.

    Args:
        modelname (str): The model to create.
        pretrained (bool, optional): If pretrained model is to be returned. Defaults to True.
        num_classes (int, optional): The number of output classes. Defaults to 2.
        dimensions (int, optional): The dimensionality of computations. Defaults to 2.

    Returns:
        model (torchvision.models.model): The created model after taking output classes into account.
    """

    assert dimensions == 2, "ImageNet_VGG only supports 2D images"

    if modelname == "vgg11":
        model = torchvision.models.vgg11(
            pretrained=pretrained,
        )
    if modelname == "vgg11_bn":
        model = torchvision.models.vgg11_bn(
            pretrained=pretrained,
        )
    if modelname == "vgg13":
        model = torchvision.models.vgg13(
            pretrained=pretrained,
        )
    if modelname == "vgg13_bn":
        model = torchvision.models.vgg13_bn(
            pretrained=pretrained,
        )
    if modelname == "vgg16":
        model = torchvision.models.vgg16(
            pretrained=pretrained,
        )
    if modelname == "vgg16_bn":
        model = torchvision.models.vgg16_bn(
            pretrained=pretrained,
        )
    if modelname == "vgg19":
        model = torchvision.models.vgg19(
            pretrained=pretrained,
        )
    if modelname == "vgg19_bn":
        model = torchvision.models.vgg19_bn(
            pretrained=pretrained,
        )
    model.classifier[6] = nn.Linear(
        in_features=model.classifier[3].out_features, out_features=num_classes
    )
    return model


def apply_activation_function(activation_function, input_tensor):
    out = input_tensor
    if not activation_function is None:
        if activation_function == F.softmax:
            out = activation_function(out, dim=1)
        else:
            out = activation_function(out)
    return out


class imagenet_vgg11(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg11, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg11",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg11_bn(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg11_bn, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg11_bn",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg13(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg13, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg13",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg13_bn(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg13_bn, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg13_bn",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg16(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg16, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg16",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg16_bn(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg16_bn, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg16_bn",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg19(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg19, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg19",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))


class imagenet_vgg19_bn(ModelBase):
    def __init__(
        self,
        parameters,
    ) -> None:
        super(imagenet_vgg19_bn, self).__init__(parameters)

        pretrained = parameters["model"].get("pretrained", True)
        self.model = create_torchvision_model(
            "vgg19_bn",
            pretrained=pretrained,
            num_classes=self.n_classes,
            dimensions=self.n_dimensions,
        )

    def forward(self, x):
        return apply_activation_function(self.final_convolution_layer, self.model(x))
